"""
Qlattice package to store parameters of RBC-UKQCD configurations.
"""

dict_params = {}

from . import p_test

from . import p_24D
from . import p_32D

from . import p_32Dfine

from . import p_24DH

from . import p_16IH2

from . import p_24IH1
from . import p_24IH2
from . import p_24IH3
from . import p_24IH4

from . import p_32IH1
from . import p_32IH2
from . import p_32IH3

from . import p_32IcoarseH1

from . import p_32IfineH

from . import p_48I
from . import p_64I
from . import p_96I

from . import p_24IH01
from . import p_24IH02
from . import p_24IH03
from . import p_24IH04
from . import p_24IH05

from . import p_32IcoarseH01
from . import p_32IcoarseH02

from . import p_32IH01

from .utils import *
