#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2023
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from qlat_utils import \
        rel_mod, \
        rel_mod_sym, \
        c_rel_mod_sqr

auto_fac_funcs_list = [
        "rel_mod",
        "rel_mod_sym",
        "c_rel_mod_sqr",
        ]
